"""

    Account class to provide basic functionality for convex network

"""

from convex_api import Account as ConvexAPIAccount

from starfish.network.account_base import AccountBase


class ConvexAccount(ConvexAPIAccount, AccountBase):
    pass



"""

    Account class to provide basic functionality for all Starfish library accounts.

"""

from abc import (
    ABC,
    abstractmethod
)

from typing import Any


class AccountBase(ABC):
    """

    Base Account class


    """

    @abstractmethod
    def sign_transaction(self, transaction: Any) -> Any:
        pass



"""

    Command 'Asset'

"""
from typing import Any

from .asset_download_command import AssetDownloadCommand
from .asset_store_command import AssetStoreCommand
from .command_base import CommandBase
from .help_command import HelpCommand


class AssetCommand(CommandBase):

    def __init__(self, sub_parser: Any = None) -> None:
        self._command_list = []
        super().__init__('asset', sub_parser)

    def create_parser(self, sub_parser: Any) -> Any:

        parser = sub_parser.add_parser(
            self._name,
            description='Tool tasks for assets',
            help='Tasks to perform on and with assets',

        )
        asset_parser = parser.add_subparsers(
            title='Asset sub command',
            description='Asset sub command',
            help='Asset sub command',
            dest='asset_command'
        )

        self._command_list = [
            AssetDownloadCommand(asset_parser),
            AssetStoreCommand(asset_parser),
            HelpCommand(asset_parser, self)
        ]
        return asset_parser

    def execute(self, args: Any, output: Any) -> Any:
        return self.process_sub_command(args, output, args.asset_command)



"""

    Command Account Send Token

"""
import logging
from typing import Any
from web3 import Web3

from starfish.network.ethereum.ethereum_account import EthereumAccount

from .command_base import CommandBase

logger = logging.getLogger(__name__)


class AccountSendTokenCommand(CommandBase):

    def __init__(self, sub_parser: Any = None) -> None:
        super().__init__('token', sub_parser)

    def create_parser(self, sub_parser: Any):

        parser = sub_parser.add_parser(
            self._name,
            description='Send tokens for an account address',
            help='Send tokens from one account to the another'
        )

        parser.add_argument(
            'address',
            help='Account address'
        )

        parser.add_argument(
            'password',
            help='Account password'
        )

        parser.add_argument(
            'keyfile',
            help='Account keyfile'
        )

        parser.add_argument(
            'to_address',
            help='Address to send data too',
        )

        parser.add_argument(
            'amount',
            help='Amount to send'
        )
        return parser

    def execute(self, args: Any, output: Any) -> None:
        if not Web3.isAddress(args.address):
            output.add_error(f'{args.address} is not an ethereum account address')
            return

        account = EthereumAccount(args.address, args.password, key_file=args.keyfile)
        amount = float(args.amount)
        to_address = args.to_address

        network = self.get_network(args.url)
        logger.debug(f'sending tokens from account {account.address} to account {to_address}')
        network.send_token(account, to_address, amount)

        balance = network.get_token_balance(account)
        output.add_line(f'Send {amount} tokens from account: {args.address} to account {to_address}')
        output.set_value('balance', balance)
        output.set_value('from_address', args.address)
        output.set_value('to_address', args.to_address)
        output.set_value('amount', amount)



"""


Contract Base

"""

from starfish.network.convex.convex_network import ConvexNetwork


class ContractBase:
    def __init__(self, convex: ConvexNetwork, name: str):
        self._convex = convex
        self._name = name
        self._version = None
        self._address = None

    def load(self, deploy_account: str):
        self._address = self.get_address(deploy_account)
        self._version = self.get_version(deploy_account)
        return (self._address, self._version)

    def get_address(self, deploy_account):
        address = self._convex.get_address(self._name, deploy_account)
        return address

    def get_version(self, deploy_account):
        address = self._address
        if address is None:
            address = self.get_address(deploy_account)

        result = self._convex.query(f'(call {address} (version))', deploy_account)
        if result and 'value' in result:
            return result['value']

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    @property
    def address(self):
        return self._address



"""

    Account class to provide basic functionality for all Starfish library accounts.

"""

import json
from typing import Any
from eth_account import Account as EthAccount
from web3 import Web3

from starfish.network.account_base import AccountBase


class EthereumAccount(AccountBase):
    """

    Account class, adds functionality for an account to be used on the starfish named network.

    :param address: address or dict of the account details
    :type address: string or dict
    :param password: password for the account
    :type password: str or None
    :param str key_data: key value data that is stored in the keyfile
    :param keyfile: keyfile containing the private encrypted key
    :type keyfile: str or None


    If the address parameter is a string then it's the account address.
    If dict then the dict can be the following format: ::

        {
            'address': 'xxxx',
            'password': 'yyyy',
        }

    """

    def __init__(self, key_data: Any, password: str) -> None:
        """init a standard account object"""
        self._address = None
        self._password = password
        self._key_data = key_data
        if isinstance(self._key_data, dict):
            self._address = Web3.toChecksumAddress(self._key_data['address'])

    @staticmethod
    def create(password: str) -> AccountBase:
        """

        Create a new account.

        You need to call `save_key_file` method to save the key information

        :return: Returns an account object
        """
        local_account = EthAccount.create(password)
        key_data = EthAccount.encrypt(local_account.key, password)
        account = EthereumAccount(key_data, password)
        return account

    @staticmethod
    def import_from_text(data, password) -> AccountBase:
        return EthereumAccount(data, password)

    @staticmethod
    def import_from_file(filename, password) -> AccountBase:
        with open(filename, 'r') as fp:
            data = json.load(fp)
            return EthereumAccount(data, password)

    def export_to_file(self, filename: str) -> None:
        """

        Save a key value to a file

        :param str filename: filename to write the key_data too.

        """
        if self._key_data:
            with open(filename, 'w') as fp:
                json.dump(self._key_data, fp)

    @property
    def export_to_text(self):
        """

        Export a key_data to json text

        """
        return json.dumps(self._key_data, sort_keys=True, indent=2)

    def is_address_equal(self, address: str) -> bool:
        """

        Compares two addresses if are equal. Both addresses are converted to checksum
        address and then compared.

        :param str address: address to compare with this object's address
        :return: True if the param address is the same is the one held in this account
        :type: bool

        >>> account = Account('0x00bd138abd70e2f00903268f3db08f2d25677c9e')
        >>> account.is_address_equal('0x00Bd138aBD70e2F00903268F3Db08f2D25677C9e')
        True
        """
        return self.as_checksum_address == Web3.toChecksumAddress(address)

    def sign_transaction(self, transaction: Any, web3: Any) -> Any:
        if self._key_data:
            secret_key = web3.eth.account.decrypt(self._key_data, self._password)
            signed = web3.eth.account.sign_transaction(transaction, secret_key)
            return signed

    @property
    def address(self) -> str:
        """

        Return the account address for this account

        :return: address
        :type: str

        >>> account = Account('0x00bd138abd70e2f00903268f3db08f2d25677c9e')
        >>> account.address
        0x00bd138abd70e2f00903268f3db08f2d25677c9e
        """
        return self._address

    @property
    def as_checksum_address(self) -> str:
        """

        Return the address as a checksum address

        :return: checksum address
        :type: str

        >>> account = Account('0x00bd138abd70e2f00903268f3db08f2d25677c9e')
        >>> account.as_checksum_address
        0x00Bd138aBD70e2F00903268F3Db08f2D25677C9e

        """
        if self._address:
            return Web3.toChecksumAddress(self._address.lower())
        return None

    @property
    def password(self) -> str:
        """

        Return the account password for this account

        :return: password
        :type: str

        >>> account.password
        secret

        """
        return self._password

    @property
    def key_data(self) -> Any:
        """

        This is the encrypted key value that contains the account private key

        """
        return self._key_data

    @property
    def is_valid(self) -> bool:
        return self._address and self._password and self._key_data

    def __str__(self) -> str:
        return f'Account: {self.address}'



"""

    Network class

    To access block chain network services.

"""

import logging

from abc import (
    ABC,
    abstractmethod
)

from starfish.network.account_base import AccountBase
from starfish.network.ddo import DDO
from starfish.network.did import is_did
from starfish.types import Authentication

logger = logging.getLogger(__name__)


class NetworkBase(ABC):

    def __init__(self, url: str):
        self._url = url

    """

    Register DID with a DDO and resolve DID to a DDO

    """
    @abstractmethod
    def register_did(self, account: AccountBase, did: str, ddo_text: str) -> bool:
        return False

    @abstractmethod
    def resolve_did(self, did: str) -> str:
        return None

    """


    Helper methods


    """

    def resolve_agent(
        self,
        agent_url_did: str,
        username: str = None,
        password: str = None,
        authentication: Authentication = None
    ) -> DDO:

        # stop circular references on import

        from starfish.agent.remote_agent import RemoteAgent

        ddo = None
        if is_did(agent_url_did):
            ddo_text = self.resolve_did(agent_url_did)
            if ddo_text:
                ddo = DDO.import_from_text(ddo_text)
            return ddo

        if not authentication:
            if username or password:
                authentication = {
                    'username': username,
                    'password': password
                }
        ddo_text = RemoteAgent.resolve_url(agent_url_did, authentication)
        if ddo_text:
            ddo = DDO.import_from_text(ddo_text)
        return ddo

    @property
    def url(self) -> str:
        return self._url



"""

    Agent Manager


"""

from starfish.agent import RemoteAgent
from starfish.agent_manager.agent_access import AgentAccess
from starfish.network.ddo import DDO

LOCAL_AGENT_NAME = '_local_agent'


class AgentManager:

    def __init__(self, http_client=None):
        """
        Create an agent manager object to resolve agents and keep a list of knwon agents.
        If the list of known agents only have a url, then the agent manager will try
        to resolve the url's to DID's, using the api call to obtain the DDO from the
        remote agents.

        """
        self._agent_access_list = []
        self._local_name = LOCAL_AGENT_NAME
        self._http_client = http_client

    def register_agents(self, agents):
        """
        Register a dict of agents and their url/did and access information.
        This allows the agent manager to resolve 'agent names'.
        So that you can pass a given name for the agent

        :param dict agents: Dict of agents that can be stored in the agent manager.

        """
        if not isinstance(agents, dict):
            raise TypeError('agents parameter must be a type dict')

        for name, agent in agents.items():
            agent_access = AgentAccess(
                name,
                url=agent.get('url', None),
                did=agent.get('did', None),
                username=agent.get('username', None),
                password=agent.get('password', None),
                token=agent.get('token', None),
            )
            self._agent_access_list.append(agent_access)

    def register_local_agent(self, ddo_text, authentication=None, local_name=None):
        """
        Register a local agent with the  agent manager object.

        :param str ddo_text: DDO string of the local agent.
        :param dict authentication: Authentication dict that can access the local agent
        :param str local_name: Optional local name to use for this agent

        """
        if isinstance(ddo_text, DDO):
            ddo_text = ddo_text.as_text

        self._agent_access_list = []
        if local_name:
            self._local_name = local_name
        agent_access = AgentAccess(self._local_name, ddo_text=ddo_text, authentication=authentication)
        self._agent_access_list.append(agent_access)

    def resolve_agent_url(self, url,  authentication=None):
        """
        Resolve an agent using ony the URL of the remote agent.

        :param str url: URL of the remote agent to resolve

        :returns: RemoteAgent object if found, else return None

        """
        ddo = AgentAccess.resolve_agent_url(url, authentication=authentication, http_client=self._http_client)
        if ddo:
            return RemoteAgent(ddo, authentication=authentication, http_client=self._http_client)

    def resolve_agent_did(self, did,  authentication=None):
        """
        Resolve an agent using ony the DID of the remote agent.

        :param str did: DID of the remote agent to resolve

        :returns: RemoteAgent object if found, else return None

        """
        ddo = AgentAccess.resolve_agent_url(did, authentication=authentication, http_client=self._http_client)
        if ddo:
            return RemoteAgent(ddo, authentication=authentication, http_client=self._http_client)

    def load_agent(self, name_did_url, authentication=None):
        """
        Loads and returns a starfish RemoteAgent that is found using the agent name/did/url.

        :param str name_did_url: name, did or url of the agent to find
        :param dict authentication: Optional authentication dict to access this server.

        :returns: RemoteAgent object for the loaded local agent.
        """

        found_agent_access = self.find_agent_access(name_did_url)
        if found_agent_access:
            return found_agent_access.load_agent(authentication=authentication, http_client=self._http_client)

    def load_ddo(self, name_did_url, authentication=None):
        """
        Loads and returns a starfish DDO that is found using the agent name/did/url.

        :param str name_did_url: name, did or url of the agent to find
        :param dict authentication: Optional authentication dict to access this server.

        :returns: RemoteAgent object for the loaded local agent.
        """

        found_agent_access = self.find_agent_access(name_did_url)
        if found_agent_access:
            return found_agent_access.ddo

    def find_agent_access(self, name_did_url, auto_resolve=True):
        """

        Finds an agent in the access records of this manager.
        If found return the agent access record

        :param str name_did_url: name/did or url of the agent to find in the access records
        :param bool auto_resolve: If True then if no agent found, try to resolve all of the agents,
            and call again with this field set too False.

        :returns: AgentAccess object if found
        """

        for agent_access in self._agent_access_list:
            if agent_access.is_match(name_did_url):
                # only return if we have a valid DDO
                if agent_access.ddo:
                    return agent_access

        if auto_resolve:
            self.resolve_access_agents()

        # call again but with no auto_resolve
        return self.find_agent_access(name_did_url, False)

    def is_agent(self, name_did_url):
        """
        Is the name/did or url in the agent access record list

        :param str name_did_url: name/did or url of the agent to find in the access records

        :returns: True if found

        """
        return self.find_agent_access(name_did_url) is not None

    def resolve_access_agents(self):
        for agent_access in self._agent_access_list:
            if agent_access.did is None and agent_access.url:
                agent_access.resolve_url(http_client=self._http_client)

    @property
    def local_agent(self):
        """
        Returns the local agent, if it has been registered

        """
        return self.load_agent(self._local_name)

    @property
    def local_agent_name(self):
        return self._local_agent_name

    @property
    def access_items(self):
        return self._agent_access_list

    @property
    def http_client(self):
        return self._http_client

    @http_client.setter
    def http_client(self, value):
        self._http_client = value
        found_agent_access = self.find_agent_access(self._local_name)
        if found_agent_access:
            found_agent_access.clear_cache()



"""

    Command 'Asset Store'

"""
from typing import Any

from starfish.agent import RemoteAgent
from .command_base import CommandBase


class AssetDownloadCommand(CommandBase):

    def __init__(self, sub_parser: Any = None) -> None:
        self._command_list = []
        super().__init__('download', sub_parser)

    def create_parser(self, sub_parser: Any) -> Any:

        parser = sub_parser.add_parser(
            self._name,
            description='Tool to download an asset',
            help='Tool to download an asset',
        )

        parser.add_argument(
            '-u',
            '--username',
            help='Optional username to access the agent'
        )

        parser.add_argument(
            '-p',
            '--password',
            help='Optional password to access the agent'
        )

        parser.add_argument(
            'asset_did',
            help='asset did of the asset to download'
        )

        parser.add_argument(
            'filename',
            nargs='?',
            help='filename to download. Defaults: to the "<asset_did>.dat" '
        )

        return parser

    def execute(self, args: Any, output: Any) -> None:

        network = self.get_network(args.url)

        ddo = network.resolve_agent(args.asset_did, username=args.username, password=args.password)
        if not ddo:
            output.add_line(f'cannot resolve asset {args.asset_did}')
            return

        authentication = None
        if args.username or args.password:
            authentication = {
                'username': args.username,
                'password': args.password
            }

        agent = RemoteAgent(ddo=ddo, authentication=authentication)
        asset = agent.download_asset(args.asset_did)
        asset_filename = args.filename
        if asset_filename is None:
            asset_filename = f'{asset.id}.dat'

        asset.save_to_file(asset_filename)
        output.add_line(f'saved asset {asset_filename}')
        output.add_line(f'asset metadata {asset.metadata}')
        output.set_value('asset_did', asset.did)
        output.set_value('filename', asset_filename)
        output.set_value('metadata', asset.metadata)



"""

    Network class

    To access block chain network services.

"""

import logging
from typing import (
    Any,
    List
)
import requests

from web3 import (
    HTTPProvider,
    Web3
)

from web3.gas_strategies.rpc import rpc_gas_price_strategy
from web3.middleware import geth_poa_middleware

from starfish.exceptions import (
    StarfishConnectionError,
    StarfishInsufficientFunds
)
from starfish.network.did import is_did
from starfish.network.ethereum.ethereum_account import EthereumAccount
from starfish.network.network_base import NetworkBase
from starfish.types import (
    AccountAddress,
    ProvenanceEventList,
    TContractBase
)


logger = logging.getLogger(__name__)

DEFAULT_PACKAGE_NAME = 'starfish.network.ethereum.contract'

NETWORK_NAMES = {
    0: 'development',
    1: 'main',
    2: 'morden',
    3: 'ropsten',
    4: 'rinkeby',
    42: 'kovan',
    77: 'POA_Sokol',
    99: 'POA_Core',
    100: 'xDai',
    1337: 'local',                  # Local private network
    8995: 'nile',                   # Ocean Protocol Public test net
    8996: 'spree',                  # Ocean Protocol local test net
    0xcea11: 'pacific'              # Ocean Protocol Public mainnet
}

CONTRACT_LIST = {
    'Network': {
        'name': 'NetworkContract',
        'has_artifact': False,
    },
    'DIDRegistry': {
        'name': 'DIDRegistryContract',
    },
    'DirectPurchase': {
        'name': 'DirectPurchaseContract'
    },
    'DatacraftToken': {
        'name': 'DatacraftTokenContract'
    },
    'Dispenser': {
        'name': 'DispenserContract'
    },
    'Provenance': {
        'name': 'ProvenanceContract',
    },

}


class EthereumNetwork(NetworkBase):
    def __init__(self, url: str, artifacts_path: str = None, load_development_contracts: bool = True) -> None:
        NetworkBase.__init__(self, url)
        self._web3 = None
        self._name = None
        self._id = None
        self._artifacts_path = artifacts_path
        self._contracts = {}
        if artifacts_path is None:
            artifacts_path = 'artifacts'

        from starfish.network.ethereum.contract.contract_manager import ContractManager

        if self._connect(self._url, artifacts_path):
            self._contract_manager = ContractManager(
                self,
                DEFAULT_PACKAGE_NAME,
                artifacts_path,
            )
            if load_development_contracts and self._name == 'local':
                self._contract_manager.load_local_artifacts_package()

    def get_contract(self, name: str) -> TContractBase:
        if name not in CONTRACT_LIST:
            raise LookupError(f'Invalid contract name: {name}')

        if name not in self._contracts:
            item = CONTRACT_LIST[name]
            self._contracts[name] = self._contract_manager.load(
                item['name'],
                artifact_filename=item.get('artifact_filename', None),
                has_artifact=item.get('has_artifact', True)
            )
        return self._contracts[name]

    """

    Account based operations


    """
    def get_ether_balance(self, account_address: AccountAddress) -> float:
        network_contract = self.get_contract('Network')
        return network_contract.get_balance(account_address)

    def get_token_balance(self, account_address: AccountAddress) -> float:
        datacraft_token_contract = self.get_contract('DatacraftToken')
        return datacraft_token_contract.get_balance(account_address)

    def request_test_tokens(self, account: EthereumAccount, amount: float) -> bool:
        dispenser_contract = self.get_contract('Dispenser')
        tx_hash = dispenser_contract.request_tokens(account, amount)
        receipt = dispenser_contract.wait_for_receipt(tx_hash)
        return receipt.status == 1

    def request_ether_from_faucet(self, account: EthereumAccount, url: str) -> None:
        data = {
            'address': account.address,
            'agent': 'server',
        }
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }
        response = requests.post(url, json=data, headers=headers)
        logger.debug(f'response {response.text} {response.status_code}')
        if response.status_code != 200:
            logger.warning(f'{response.status_code} {response.text}')

    """

    Send ether and tokens to another account

    """
    def send_ether(self, account: EthereumAccount, to_account_address: AccountAddress, amount: float) -> bool:
        network_contract = self.get_contract('Network')

        account_balance = self.get_ether_balance(account)
        if account_balance < amount:
            raise StarfishInsufficientFunds(f'The account has insufficient funds to send {amount} tokens')

        tx_hash = network_contract.send_ether(account, to_account_address, amount)
        receipt = network_contract.wait_for_receipt(tx_hash)
        return receipt.status == 1

    def send_token(self, account: EthereumAccount, to_account_address: AccountAddress, amount: float) -> bool:
        datacraft_token_contract = self.get_contract('DatacraftToken')

        account_balance = self.get_token_balance(account)
        if account_balance < amount:
            raise StarfishInsufficientFunds(f'The account has insufficient funds to send {amount} tokens')

        tx_hash = datacraft_token_contract.transfer(account, to_account_address, amount)
        receipt = datacraft_token_contract.wait_for_receipt(tx_hash)
        return receipt.status == 1

    """

    Send tokens (make payment) with logging

    """
    def send_token_and_log(
        self,
        account: EthereumAccount,
        to_account_address: AccountAddress,
        amount: float,
        reference_1: str = None,
        reference_2: str = None
    ) -> bool:
        datacraft_token_contract = self.get_contract('DatacraftToken')
        direct_contract = self.get_contract('DirectPurchase')

        account_balance = self.get_token_balance(account)
        if account_balance < amount:
            raise StarfishInsufficientFunds(f'The account has insufficient funds to send {amount} tokens')

        tx_hash = datacraft_token_contract.approve_transfer(
            account,
            direct_contract.address,
            amount
        )
        receipt = datacraft_token_contract.wait_for_receipt(tx_hash)
        if receipt and receipt.status == 1:
            tx_hash = direct_contract.send_token_and_log(
                account,
                to_account_address,
                amount,
                reference_1,
                reference_2
            )
            receipt = direct_contract.wait_for_receipt(tx_hash)
            if receipt and receipt.status == 1:
                return True
        return False

    def is_token_sent(
        self,
        from_account_address: AccountAddress,
        to_account_address: AccountAddress,
        amount: float,
        reference_1: str = None,
        reference_2: str = None
    ) -> bool:
        direct_contract = self.get_contract('DirectPurchase')

        is_sent = direct_contract.check_is_paid(
            from_account_address,
            to_account_address,
            amount,
            reference_1,
            reference_2
        )
        return is_sent

    """

    Register Provenance

    """
    def register_provenace(self, account: EthereumAccount, asset_id: str) -> bool:
        provenance_contract = self.get_contract('Provenance')
        tx_hash = provenance_contract.register(account, asset_id)
        receipt = provenance_contract.wait_for_receipt(tx_hash)
        return receipt.status == 1

    def get_provenace_event_list(self, asset_id: str) -> ProvenanceEventList:
        provenance_contract = self.get_contract('Provenance')
        return provenance_contract.get_event_list(asset_id)

    """

    Register DID with a DDO and resolve DID to a DDO

    """
    def register_did(self, account: EthereumAccount, did: str, ddo_text: str) -> bool:
        did_registry_contract = self.get_contract('DIDRegistry')
        tx_hash = did_registry_contract.register(account, did, ddo_text)
        receipt = did_registry_contract.wait_for_receipt(tx_hash)
        return receipt.status == 1

    def resolve_did(self, did: str) -> str:
        ddo_text = None
        if did and is_did:
            did_registry_contract = self.get_contract('DIDRegistry')
            ddo_text = did_registry_contract.get_value(did)
        return ddo_text

    @property
    def contract_names(self) -> List[str]:
        return CONTRACT_LIST.keys()

    @property
    def name(self) -> str:
        return self._name

    @property
    def network_id(self) -> int:
        return self._id

    @property
    def web3(self) -> Any:
        return self._web3

    @staticmethod
    def find_network_name_from_id(network_id: int) -> str:
        if network_id in NETWORK_NAMES:
            return NETWORK_NAMES[network_id]
        return NETWORK_NAMES[0]

    def _connect(self, url: str, artifacts_path: str) -> bool:
        self._url = url
        self._web3 = Web3(HTTPProvider(url))
        if self._web3:

            try:
                self._id = int(self._web3.net.version)
                self._name = EthereumNetwork.find_network_name_from_id(self._id)
            except requests.exceptions.ConnectionError as e:
                raise StarfishConnectionError(e)

            if self._name == 'local':
                # inject the poa compatibility middleware to the innermost layer
                self._web3.middleware_onion.inject(geth_poa_middleware, layer=0)

            logger.info(f'connected to the {self._name} network')
            self._web3.eth.setGasPriceStrategy(rpc_gas_price_strategy)
            return True
        return False

    def __str__(self) -> str:
        return f'Account: {self.address}'



"""

    DIDRegistry Contract

"""
import logging

from starfish.network.ethereum.contract.contract_base import ContractBase
from starfish.network.ethereum.ethereum_account import EthereumAccount
from starfish.types import AccountAddress

logger = logging.getLogger(__name__)

CONTRACT_NAME = 'Network'


class NetworkContract(ContractBase):
    """

    Class representing a network contract, that does not access any actual underling contract, but does network based
    tasks.

    """

    def __init__(self) -> None:
        ContractBase.__init__(self, CONTRACT_NAME)

    def get_balance(self, account_address: AccountAddress) -> float:

        address = self.get_account_address(account_address)
        amount_wei = self._web3.eth.getBalance(address, block_identifier='latest')
        return self.to_ether(amount_wei)

    def send_ether(self, account: EthereumAccount, to_account_address: AccountAddress, amount: float) -> str:

        amount_wei = self.to_wei(amount)
        to_address = self.get_account_address(to_account_address)

        gas_transact = {
            'from': account.address,
            'to': to_address,
            'value': amount_wei,
        }
        transaction = {
            'from': account.address,
            'to':  to_address,
            'value': amount_wei,
            'gas': self._web3.eth.estimateGas(gas_transact),
            'gasPrice':  self.get_gas_price(account.address),
            'nonce': self.get_nonce(account.address),
        }
        signed = account.sign_transaction(transaction, self._web3)
        tx_hash = None
        if signed:
            tx_hash = self._web3.eth.sendRawTransaction(signed.rawTransaction)

        return tx_hash



"""

    Command 'Agent'

"""
from typing import Any

from .agent_register_command import AgentRegisterCommand
from .agent_resolve_command import AgentResolveCommand
from .command_base import CommandBase
from .help_command import HelpCommand


class AgentCommand(CommandBase):

    def __init__(self, sub_parser: Any = None) -> None:
        self._command_list = []
        super().__init__('agent', sub_parser)

    def create_parser(self, sub_parser: Any) -> Any:

        parser = sub_parser.add_parser(
            self._name,
            description='Tool tasks for agents',
            help='Tasks to perform on agents',

        )
        agent_parser = parser.add_subparsers(
            title='Agent sub command',
            description='Agent sub command',
            help='Agent sub command',
            dest='agent_command'
        )

        self._command_list = [
            AgentRegisterCommand(agent_parser),
            AgentResolveCommand(agent_parser),
            HelpCommand(agent_parser, self)
        ]
        return agent_parser

    def execute(self, args: Any, output: Any) -> Any:
        return self.process_sub_command(args, output, args.agent_command)



"""

    Command Base Class

"""

from abc import (
    ABC,
    abstractmethod
)
from typing import Any

from starfish.network.ethereum.ethereum_network import EthereumNetwork

DEFAULT_NETWORK_URL = 'http://localhost:8545'

NETWORK_NAMES = {
    'local': {
        'description': 'Spree network running on a local barge',
        'url': 'http://localhost:8545',
        'faucet_account': ['0x068Ed00cF0441e4829D9784fCBe7b9e26D4BD8d0', 'secret'],
    },
    'spree': {
        'description': 'Spree network running on a local barge',
        'url': 'http://localhost:8545',
        'faucet_account': ['0x068Ed00cF0441e4829D9784fCBe7b9e26D4BD8d0', 'secret'],
    },
    'nile': {
        'description': 'Nile network access to remote network node',
        'url': 'https://nile.dev-ocean.com',
        'faucet_url': 'https://faucet.nile.dev-ocean.com/faucet',
    },
    'pacific': {
        'description': 'Pacific network access to remote network node',
        'url': 'https://pacific.oceanprotocol.com',
        'faucet_url': 'https://faucet.oceanprotocol.com/faucet',
    },
    'duero': {
        'description': 'Duero network access to remote network node',
        'url': 'https://duero.dev-ocean.com',
        'faucet_url': 'https://faucet.duero.dev-ocean.com/faucet',
    },
    'host': {
        'description': 'Local node running on barge',
        'url': 'http://localhost:8545',
    }
}


class CommandBase(ABC):
    def __init__(self, name: str, sub_parser: Any = None) -> None:
        self._name = name
        self._sub_parser = sub_parser
        if sub_parser:
            self.create_parser(sub_parser)

    def is_command(self, name: str) -> bool:
        return self._name == name

    def get_network(self, url: str, default_url: str = None, load_development_contracts: bool = True) -> Any:
        if url is None:
            url = default_url
        if url is None:
            url = 'http://localhost:8545'
        network = EthereumNetwork(url, load_development_contracts=load_development_contracts)

        return network

    def get_network_setup(self, name: str) -> Any:
        result = None
        if name in NETWORK_NAMES:
            result = NETWORK_NAMES[name]
        return result

    def process_sub_command(self, args: Any, output: Any, command: str) -> None:
        is_found = False
        for command_item in self._command_list:
            if command_item.is_command(command):
                command_item.execute(args, output)
                is_found = True
                break

        if not is_found:
            self.print_help()

    def print_help(self) -> None:
        self._sub_parser.choices[self._name].print_help()

    @abstractmethod
    def create_parser(self, sub_parser: Any) -> Any:
        pass

    @abstractmethod
    def execute(self, args: Any, output: Any) -> None:
        pass

    @property
    def name(self) -> str:
        return self._name



import secrets

from eth_account import Account as EthAccount
from web3 import Web3

class UnitTestAccount():

    def __init__(self, password=None):
        self.password = secrets.token_hex(32)
        if password:
            self.password = password
        local_account = EthAccount.create(password)
        self.key_data = EthAccount.encrypt(local_account.key, password)
        self.address = Web3.toChecksumAddress(self.key_data['address'])




"""

    Command Account Get Token

"""
import logging
from typing import Any
from web3 import Web3

from starfish.network.ethereum.ethereum_account import EthereumAccount

from .command_base import CommandBase

logger = logging.getLogger(__name__)


DEFAULT_AMOUNT = 10


class AccountGetTokenCommand(CommandBase):

    def __init__(self, sub_parser: Any = None) -> None:
        super().__init__('token', sub_parser)

    def create_parser(self, subparser: Any) -> Any:

        parser = subparser.add_parser(
            self._name,
            description='Get tokens for an account address',
            help='Get some token from the development network'
        )

        parser.add_argument(
            'address',
            help='Account address'
        )

        parser.add_argument(
            'password',
            help='Account password'
        )

        parser.add_argument(
            'keyfile',
            help='Account keyfile'
        )

        parser.add_argument(
            'amount',
            nargs='?',
            default=DEFAULT_AMOUNT,
            help=f'Amount to request. Default {DEFAULT_AMOUNT}'
        )
        return parser

    def execute(self, args: Any, output: Any) -> None:
        if not Web3.isAddress(args.address):
            output.add_error(f'{args.address} is not an ethereum account address')
            return

        account = EthereumAccount(args.address, args.password, key_file=args.keyfile)

        network = self.get_network(args.url)
        network.request_test_tokens(account, int(args.amount))

        logger.debug(f'requesting tokens for account {account.address}')
        balance = network.get_token_balance(account)
        output.add_line(f'Get {args.get_command} for account: {args.address} balance: {balance}')
        output.set_value('balance', balance)
        output.set_value('address', args.address)



"""

    Types used for typing


"""
import sys

from datetime import datetime
from typing import (
    Any,
    List,
    Tuple,
    TypeVar,
    Union
)

# TypeDict is only available in python3.8
if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

TAccountBase = TypeVar('AccountBase')
TAsset = TypeVar('Asset')
TAssetBase = TypeVar('AssetBase')
TBundleAsset = TypeVar('BundleAsset')
TDataAsset = TypeVar('DataAsset')
TOperationAsset = TypeVar('OperationAsset')
TRemoteDataAsset = TypeVar('RemoteDataAsset')
TContractBase = TypeVar('ContractBase')
TContractManager = TypeVar('ContractManager')
TDDO = TypeVar('DDO')
TJobBase = TypeVar('JobBase')
TListing = TypeVar('Listing')
TListingBase = TypeVar('ListingBase')
TNetworkBase = TypeVar('NetworkBase')
TRemoteAgent = TypeVar('RemoteAgent')


class Authentication(TypedDict):
    username: str
    password: str
    token: str


class AccountInitParams(TypedDict):
    address: str
    password: str
    key_value: Any
    key_file: str


class AgentItem(TypedDict):
    url: str
    did: str
    authentication: Authentication


class DIDParts(TypedDict):
    method: str
    id: str
    path: str
    fragment: str
    id_hex: str


class ProvenanceEvent(TypedDict):
    asset_id: str
    account: str
    timestamp: datetime


class ListingData(TypedDict):
    trust_level: int
    userid: str
    assetid: str
    agreement: str
    ctime: str
    status: str
    id: str
    info: Any
    utime: str


AccountAddressOrDict = Union[AccountInitParams, str, Tuple[str, str, str]]

AccountAddress = Union[TAccountBase, str]

ProvenanceEventList = List[ProvenanceEvent]



"""

Surfer Agent class to provide basic functionality for Ocean Agents

In starfish-java, this is named as `RemoteAgent`

"""

import logging
import time
from typing import (
    Any,
    Generic,
    List,
    Union
)
from urllib.parse import urljoin


from eth_utils import remove_0x_prefix
from mongoquery import Query

from starfish.agent.agent_base import AgentBase
from starfish.asset import (
    DataAsset,
    OperationAsset,
    create_asset_from_metadata_text,
    is_asset_hash_valid
)
from starfish.exceptions import (
    StarfishAssetInvalid,
    StarfishRemoteAgentInvalidAccess
)
from starfish.job import Job
from starfish.listing import Listing
from starfish.middleware.agent.remote_agent_adapter import RemoteAgentAdapter
from starfish.network.account_base import AccountBase
from starfish.network.ddo import DDO
from starfish.network.did import (
    decode_to_asset_id,
    did_parse
)
from starfish.network.network_base import NetworkBase
from starfish.types import (
    Authentication,
    ListingData,
    TAsset,
    TRemoteAgent
)

SUPPORTED_SERVICES = {
    'meta': 'DEP.Meta.v1',
    'storage': 'DEP.Storage.v1',
    'invoke': 'DEP.Invoke.v1',
    'market': 'DEP.Market.v1',
    'trust': 'DEP.Trust.v1',
    'auth': 'DEP.Auth.v1',
    'collection': 'DEP.Collection.v1',
}


logger = logging.getLogger(__name__)


class RemoteAgent(AgentBase, Generic[TRemoteAgent]):
    """

    Remote Agent class allows to register, list, purchase and consume assets.

    :param ddo: Optional ddo of the remote agent, if not provided the agent
        will automatically get the DDO from the network based on the DID.

    :param dict authentication: authentication is a dict of values to allow for authentication access to
        a remote agent.
        Currently the following values are supported:

            username
            password
            token

    """
    service_types = SUPPORTED_SERVICES

    def __init__(self, ddo: DDO, authentication: Authentication = None, http_client: Any = None) -> None:
        self._authentication = authentication

        if isinstance(ddo, str):
            ddo = DDO.import_from_text(ddo)
        elif isinstance(ddo, DDO):
            pass
        else:
            raise ValueError(f'Unknown ddo type {ddo}')
        AgentBase.__init__(self, ddo)

        self._adapter = RemoteAgentAdapter(http_client)

    @staticmethod
    def load(
        agent_did_or_url: str,
        network: NetworkBase = None,
        username: str = None,
        password: str = None,
        authentication: Authentication = None,
        http_client: Any = None
    ) -> TRemoteAgent:
        """

        Load a remote agent using a did/url. The Agent must have a DDO service available.
        If a DID is passed you also need to provide a network object to resolve the did
        For the RemoteAgent object to resolve the type of services available.

        :param str agent_did_or_url DID or URL of the agent to load.
        :param Authentiaciton authentication Optional Authenentication object to use to access the new agent.
        :param http_client: HTTP Client libray to use to make requests, this defaults to requests library.

        :return: A RemoteAgent Object or None if the agent cannot be accessed or cannot be found at the url.

        """

        if not authentication:
            if username or password:
                authentication = {
                    'username': username,
                    'password': password
                }
        ddo_text = None
        if network:
            ddo_text = network.resolve_agent(agent_did_or_url, authentication=authentication)
        else:
            ddo_text = RemoteAgent.resolve_url(agent_did_or_url, authentication=authentication, http_client=http_client)

        if ddo_text:
            return RemoteAgent(ddo_text, authentication=authentication, http_client=http_client)

        return None

    @staticmethod
    def register(
        network: NetworkBase,
        register_account: AccountBase,
        ddo: DDO,
        authentication: Authentication = None,
        http_client: Any = None
    ) -> TRemoteAgent:
        """
        Register the agent on the network.

        :param network: Network to use to register on the block chain
        :param register_account: Account object to use to register the agent ddo
        :param ddo: DDO object to use to register on the network
        :param authentication: Authentication data needed to access this agent
        :param http_client: HTTP Client libray to use to make requests, this defaults to requests library.

        :return: RemoteAgent object that has been registered on the network
        """
        network.register_did(register_account, ddo.did, ddo.as_text)
        return RemoteAgent(ddo.as_text, authentication=authentication, http_client=http_client)

    def register_asset(self, asset: TAsset, create_provenance: bool = False) -> TAsset:
        """

        Register an asset with Surfer

        :param asset: asset object to register
        :type asset: :class:`.DataAsset` object to register
        :param bool create_provenance: If set to True, then set the provenance metadata for this asset


        :return: A :class:`.AssetBase` object that has been registered, if failure then return None.
        :type: :class:`.AssetBase` class

        For example::

            asset = DataAsset.create('test data asset', 'Some test data')
            listing_data = { 'price': 3.457, 'description': 'my data is for sale' }
            agent = SurferAgent(ddo)
            asset = agent.register_asset(asset, True)
            print(f'Asset DID is {asset.did}')

        """

        if self._ddo is None:
            raise ValueError('The agent must have a valid ddo')

        url = self.get_endpoint('meta')
        authorization_token = self.get_authorization_token()
        register_data = self._adapter.register_asset(asset.metadata_text, url, authorization_token)
        if register_data:
            asset_id = register_data['asset_id']
            if asset.asset_id != asset_id:
                raise ValueError(
                    f' calculated asset_id {asset.asset_id} is not the same as the agent generated asset_id {asset_id}'
                )
            did = f'{self._ddo.did}/{asset_id}'
            asset.set_did(did)
        return asset

    def create_listing(self, listing_data: ListingData, asset_did: str) -> Listing:
        """

        Create a listing on the market place for this asset

        :param dict listing_data:  Listing inforamiton to give for this asset
        :param str asset_did: asset DID to assign to this listing
        :return: A new :class:`.Listing` object that has been registered, if failure then return None.
        :type: :class:`.Listing` class

        For example::

            asset = DataAsset.create('test data asset', 'Some test data')
            agent = SurferAgent()
            asset = agent.register_asset(asset, account)

            listing_data = { 'price': 3.457, 'description': 'my data is for sale' }
            listing = agent.create_listing(listing_data, asset)
            if listing:
                print(f'registered my listing asset for sale with the did {listing.did}')


        """

        if not isinstance(listing_data, dict):
            raise ValueError('You must provide a dict as the listing data')

        url = self.get_endpoint('market')
        authorization_token = self.get_authorization_token()

        asset_id = decode_to_asset_id(asset_did)
        data = self._adapter.create_listing(listing_data, asset_id, url, authorization_token)
        listing = Listing(self, data['id'], asset_did, data)
        return listing

    def update_listing(self, listing: Listing) -> bool:
        """

        Update the listing to the agent server.

        :param listing: Listing object to update
        :type listing: :class:`.Listing` class

        """
        url = self.get_endpoint('market')
        authorization_token = self.get_authorization_token()

        return self._adapter.update_listing(listing.listing_id, listing.data, url, authorization_token)

    def validate_asset(self, asset: TAsset) -> bool:
        """

        Validate an asset

        :param asset: Asset to validate.
        :return: True if the asset is valid
        """
        pass

    def upload_asset(self, asset: TAsset) -> bool:

        if not isinstance(asset, DataAsset):
            raise TypeError('Only DataAsset is supported')

        if not asset.data:
            raise ValueError('No data to upload')

        url = self.get_endpoint('storage')
        authorization_token = self.get_authorization_token()
        asset_id = remove_0x_prefix(asset.asset_id)

        return self._adapter.upload_asset_data(asset_id, asset.data, url, authorization_token)

    def download_asset(self, asset_did_id: str) -> TAsset:
        """
        Download an asset

        :param str asset_did_id: Asset id or asset did to download

        :return: an Asset
        :type: :class:`.DataAsset` class

        """

        url = self.get_endpoint('storage')
        authorization_token = self.get_authorization_token()

        asset_id = decode_to_asset_id(asset_did_id)
        if not asset_id:
            raise ValueError(f'{asset_did_id} is not an asset id or asset did')

        data = self._adapter.download_asset(asset_id, url, authorization_token)
        store_asset = self.get_asset(asset_id)
        asset = DataAsset(
            store_asset.metadata_text,
            data=data
        )
        asset.set_did(store_asset.did)
        return asset

    def get_listing(self, listing_id: str) -> Listing:
        """
        Return an listing on the listings id.

        :param str listing_id: Id of the listing.

        :return: a listing object
        :type: :class:`.Asset` class

        """
        listing = None
        url = self.get_endpoint('market')
        authorization_token = self.get_authorization_token()

        data = self._adapter.get_listing(listing_id, url, authorization_token)
        if data:
            asset_id = data['assetid']
            url = self.get_endpoint('meta')
            read_metadata = self._adapter.read_metadata(asset_id, url, authorization_token)
            if read_metadata:
                asset = self._create_asset_from_read(asset_id, read_metadata)
                listing = Listing(self, data['id'], asset, data)
        return listing

    def get_asset(self, asset_did_id: str) -> TAsset:
        """

        This is for compatability for Surfer calls to get an asset directly from Surfer

        :param str asset_did_id: Asset DID or asset Id of the asset to read or did/asset_id

        :return: an asset class
        :type: :class:`.AssetBase` class

        """
        asset = None
        asset_id = decode_to_asset_id(asset_did_id)
        url = self.get_endpoint('meta')
        authorization_token = self.get_authorization_token()
        read_metadata = self._adapter.read_metadata(asset_id, url, authorization_token)
        if read_metadata:
            asset = self._create_asset_from_read(asset_id, read_metadata)
        return asset

    def get_listings(self) -> List[ListingData]:
        """
        Returns all listings

        :return: List of listing objects

        """
        listings = []
        url = self.get_endpoint('market')
        authorization_token = self.get_authorization_token()
        listings_data = self._adapter.get_listings(url, authorization_token)
        if listings_data:
            for data in listings_data:
                asset_id = data['assetid']
                url = self.get_endpoint('meta')
                read_metadata = self._adapter.read_metadata(asset_id, url, authorization_token)
                if read_metadata:
                    asset = self._create_asset_from_read(asset_id, read_metadata)
                    listing = Listing(self, data['id'], asset, data)
                    listings.append(listing)
        return listings

    def get_job(self, job_id: str) -> Job:
        """

        Get a job from the invoke service ( koi )

        :param str job_id: Id of the job to get

        :return: a job class
        :type: :class:`starfish.job.Job` class

        """
        job = None
        url = self.get_endpoint('invoke', 'jobs')
        authorization_token = self.get_authorization_token()
        data = self._adapter.get_job(job_id, url, authorization_token)
        if data:
            status = data.get('status', None)
            outputs = data.get('outputs', None)
            job = Job(job_id, status, outputs)
        return job

    def job_wait_for_completion(self, job_id: str, timeout_seconds: int = 60, sleep_seconds: int = 1) -> Union[Job, bool]:
        """

        Wait for a job to complete, with optional timeout seconds.

        :param int job_id: Job id to wait for completion
        :param int timeout_seconds: optional time in seconds to wait for job to complete, defaults to 60 seconds
        :param int sleep_seconds: optional number of seconds to sleep between polling the job status, == 0 no sleeping

        :return: Job object if finished, else False for timed out
        :type: :class:`starfish.job.Job` class or False

        """
        timeout_time = time.time() + timeout_seconds
        while timeout_time > time.time():
            job = self.get_job(job_id)
            if job.is_finished:
                return job
            if sleep_seconds > 0:
                time.sleep(sleep_seconds)
        return False

    def search_listings(self, text: str, sort: Any = None, offset: int = 100, page: int = 0) -> List[ListingData]:
        """

        Search the off chain storage for an asset with the givien 'text'

        :param str text: Test to search all metadata items for.
        :param sort: sort the results ( defaults: None, no sort).
        :type sort: str or None
        :param int offset: Return the result from with the maximum record count ( defaults: 100 ).
        :param int page: Returns the page number based on the offset.

        :return: a list of assets objects found using the search.
        :type: list of DID strings

        For example::

            # return the 300 -> 399 records in the search for the text 'weather' in the metadata.
            my_result = agent.search_registered_assets('weather', None, 100, 3)

        """
        # TODO: implement search listing in remote
        pass

    def purchase_asset(self, listing: Any, AccountBase: Any, purchase_id: str = None, status: str = None,
                       info: Any = None, agreement: Any = None) -> bool:
        """

        Purchase an asset using it's listing and an account.

        :param listing: Listing to use for the purchase.
        :type listing: :class:`.Listing`
        :param account: Ocean account to purchase the asset.
        :type account: :class:`.Account` object to use for registration.
        :param purchase_id: purchase id (optional)
        :type purchase_id: str or None
        :param status: purchase status (optional)
        :type status: str or None
        :param info: purchase info (optional)
        :type info: dict or None
        :param agreement: purchase agreement (optional)
        :type agreement: dict or None

        """
        purchase = {'listingid': listing.listing_id}
        if purchase_id:
            purchase['id'] = purchase_id
        if status:
            purchase['status'] = status
        if info:
            purchase['info'] = info
        if agreement:
            purchase['agreement'] = agreement
        url = self.get_endpoint('market')
        authorization_token = self.get_authorization_token()
        return self._adapter.purchase_asset(purchase, url, authorization_token)

    def is_access_granted_for_asset(self, asset: Any, account: Any, purchase_id: str = None) -> bool:
        """

        Check to see if the account and purchase_id have access to the assed data.


        :param asset: Asset to check for access.
        :type asset: :class:`.Asset` object
        :param account: Ocean account to purchase the asset.
        :type account: :class:`.Account` object to use for registration.
        :param str purchase_id: purchase id that was used to purchase the asset.

        :return: True if the asset can be accessed and consumed.
        :type: boolean
        """
        return False

    def get_asset_purchase_ids(self, asset: Any) -> Any:
        """

        Returns as list of purchase id's that have been used for this asset

        :param asset: Asset to return purchase details.
        :type asset: :class:`.Asset` object

        :return: list of purchase ids
        :type: list

        """
        return []

    def purchase_wait_for_completion(self, asset: Any, account: Any, purchase_id: str, timeoutSeconds: int) -> bool:
        """

            Wait for completion of the purchase

            TODO: issues here...
            + No method as yet to pass back paramaters and values during the purchase process
            + We assume that the following templates below will always be used.

        """
        pass

    def consume_asset(self, listing: Any, account: Any, purchase_id: str) -> bool:
        """
        Consume the asset and download the data. The actual payment to the asset
        provider will be made at this point.

        :param listing: Listing that was used to make the purchase.
        :type listing: :class:`.Listing`
        :param str purchase_id: purchase id that was used to purchase the asset.
        :param account: Ocean account that was used to purchase the asset.
        :type account: :class:`.Account` object to use for registration.
        :param str download_path: path to store the asset data.

        :return: True if the asset has been consumed and downloaded
        :type: boolean

        """
        return False

    def invoke(self, asset: TAsset, inputs: Any = None, is_async: bool = False) -> Any:
        """

        Call an operation asset with inputs to execute a remote call.

        :param asset: Operation asset to use for this invoke call
        :type asset: :class:`.OperationAsset`
        :param dict inputs: Inputs to send to the invoke call

        :return: Return a dict of the result.
        :type: dict


        """
        if not isinstance(asset, OperationAsset):
            raise ValueError('Asset is not a OperationAsset')

        mode_type = 'async' if is_async else 'sync'
        if not asset.is_mode(mode_type):
            raise TypeError(f'This operation asset does not support {mode_type}')
        if not inputs:
            inputs = {}

        url = self.get_endpoint('invoke', mode_type)
        authorization_token = self.get_authorization_token()
        response = self._adapter.invoke(remove_0x_prefix(asset.asset_id), inputs, url, authorization_token)
        return response

    def get_collection_items(self, name=None):
        url = self.get_endpoint('collection')
        if not url:
            return None
        authorization_token = self.get_authorization_token()
        response = self._adapter.get_collection_items(url, name, authorization_token)
        return response

    def add_collection_items(self, name, asset_list):
        if not isinstance(asset_list, (tuple, list)):
            raise ValueError('You must pass a list of asset ids to add to the collection')
        url = self.get_endpoint('collection')
        if not url:
            return None
        authorization_token = self.get_authorization_token()
        response = self._adapter.add_collection_items(url, name, asset_list, authorization_token)
        return response

    def remove_collection_items(self, name, asset_list):
        if not isinstance(asset_list, (tuple, list)):
            raise ValueError('You must pass a list of asset ids to remove from the collection')
        url = self.get_endpoint('collection')
        if not url:
            return None
        authorization_token = self.get_authorization_token()
        response = self._adapter.remove_collection_items(url, name, asset_list, authorization_token)
        return response

    def get_authorization_token(self) -> str:

        if self._authentication and 'token' in self._authentication:
            if self._authentication['token']:
                return self._authentication['token']

        try:
            url = self.get_endpoint('auth', 'token')
        except ValueError:
            url = None
        token = None
        if url and self._authentication and 'username' in self._authentication:
            token = self._adapter.get_authorization_token(
                self._authentication['username'],
                self._authentication.get('password', ''),
                url
            )
            if token is None:
                raise StarfishRemoteAgentInvalidAccess(f'Unable to obtain a token from {url}')

        return token

    def get_metadata_list(self) -> Any:
        url = self.get_endpoint('meta')
        authorization_token = self.get_authorization_token()
        return self._adapter.get_metadata_list(url, authorization_token)

    def search_asset(self, filter_values: Any) -> List[str]:
        """
            :param dict filter_values: can be a dict or a mongo type query
            :returns: a list of asset_ids found
        """

        if not isinstance(filter_values, dict):
            raise TypeError('Filter values must be a type dict')

        asset_list = self.get_metadata_list()
        filter_query = Query(filter_values)
        result = []
        if asset_list:
            for asset_id, metadata in asset_list.items():
                if filter_query.match(metadata):
                    result.append(asset_id)
        return result

    @property
    def ddo(self) -> DDO:
        """

        Return the registered DDO for this agent

        :return: DDO registered for this agent
        :type: :class:`.DDO`
        """
        return self._ddo

    @staticmethod
    def is_did_valid(did: str) -> bool:
        """
        Checks to see if the DID string is a valid DID for this type of address for an asset.
        This method only checks the syntax of the DID, it does not resolve the DID
        to see if it is assigned to a valid Asset.

        :param str did: DID string to check to see if it is in a valid format.

        :return: True if the DID is in the format 'did:dep:xxxxx/yyyy'
        :type: boolean
        """
        data = did_parse(did)
        return data['path'] and data['id_hex']

    def _create_asset_from_read(self, asset_id: str, read_metadata: Any) -> Any:
        # check the hash of the reading asset
        asset_id = remove_0x_prefix(asset_id)
        if not is_asset_hash_valid(asset_id, read_metadata['hash']):
            raise StarfishAssetInvalid(f' asset {asset_id} is not valid')

        did = f'{self._ddo.did}/{asset_id}'
        metadata_text = read_metadata['metadata_text']
        asset = create_asset_from_metadata_text(metadata_text)
        asset.set_did(did)
        return asset

    def is_service(self, name):
        service_type = RemoteAgent.find_supported_service_type(name)
        if service_type is None:
            message = f'unknown surfer endpoint service name or type: {name}'
            logger.error(message)
            raise ValueError(message)
        service = self._ddo.get_service_type(service_type)
        return service is not None

    def get_endpoint(self, name: str, uri: str = None) -> str:
        """return the endpoint based on the name of the service or service type"""

        service_type = RemoteAgent.find_supported_service_type(name)
        if service_type is None:
            message = f'unknown surfer endpoint service name or type: {name}'
            logger.error(message)
            raise ValueError(message)

        endpoint = None
        if self._ddo:
            service_item = self._ddo.get_service_type(service_type)
            if not service_item:
                message = f'unable to find surfer endpoint service type {service_type}'
                logger.error(message)
                raise ValueError(message)
            endpoint = service_item['serviceEndpoint']
        if not endpoint:
            message = f'unable to find surfer endpoint for {name} = {service_type}'
            logger.error(message)
            raise ValueError(message)
        if uri:
            endpoint = urljoin(endpoint + '/', uri)
        return endpoint

    @property
    def adapter(self):
        return self._adapter

    @property
    def http_client(self):
        return self._adapter.http_client

    @http_client.setter
    def http_client(self, value):
        """Set the http client to something other than the default `requests`"""
        self._adapter.http_client = value

    @staticmethod
    def resolve_url(url: str, authentication: Authentication = None, http_client: Any = None) -> DDO:
        """

        Resolves the remote agent ddo using the url of the agent

        :param str url: url of the remote agent
        :param Authenentication: Optional authentication object to access the agent
        :param http_client: HTTP Client libray to use to make requests, this defaults to requests library.

        :return dict: DDO or None if not found
        """

        ddo = None
        if url:
            adapter = RemoteAgentAdapter(http_client)
            token = None
            # try to get a token from the agent using the username/password, or from the authentication dict
            if authentication:
                if 'username' in authentication and authentication['username']:
                    token_url = urljoin(f'{url}/', 'api/v1/auth/token')
                    token = adapter.get_authorization_token(
                        authentication['username'],
                        authentication.get('password', ''),
                        token_url
                    )
                elif 'token' in authentication and authentication['token']:
                    token = authentication['token']
            ddo = adapter.get_ddo(url, token)
        return ddo

    @staticmethod
    def find_supported_service_type(search_name_type: str) -> Any:
        """ return the supported service record if the name or service type is found
        else return None """
        for name, service_type in SUPPORTED_SERVICES.items():
            if service_type == search_name_type:
                return service_type
            if name == search_name_type:
                return service_type
        return None



"""
    starfish-ddo-registry contract

"""

from eth_utils import add_0x_prefix

from starfish.network.convex.contract.contract_base import ContractBase
from starfish.network.convex.convex_account import ConvexAccount
from starfish.network.convex.convex_network import ConvexNetwork
from starfish.types import AccountAddress


class ProvenanceContract(ContractBase):

    def __init__(self, convex: ConvexNetwork):
        ContractBase.__init__(self, convex, 'starfish-provenance')

    def register(self, asset_id: str, account: ConvexAccount):
        command = f'(call {self.address} (register {add_0x_prefix(asset_id)}))'
        result = self._convex.send(command, account)
        if result and 'value' in result:
            return {
                'timestamp': result['value']['timestamp'],
                'asset_id': add_0x_prefix(result['value']['asset-id']),
                'owner': add_0x_prefix(result['value']['owner']),
            }
        return result

    def event_list(self, asset_id: str, account_address: AccountAddress):
        command = f'(call {self.address} (event-list {add_0x_prefix(asset_id)}))'
        if isinstance(account_address, str):
            address = account_address
        else:
            address = account_address.address
        result = self._convex.query(command, address)
        if result and 'value' in result:
            return ProvenanceContract.convert_event_list(result['value'])
        return result

    def event_owner(self, account_address: AccountAddress):
        if isinstance(account_address, str):
            address = account_address
        else:
            address = account_address.address
        command = f'(call {self.address} (event-owner {add_0x_prefix(address)}))'
        result = self._convex.query(command, address)
        if result and 'value' in result:
            return ProvenanceContract.convert_event_list(result['value'])
        return result

    @staticmethod
    def convert_event_list(items):
        event_list = []
        for item in items:
            event_list.append({
                'timestamp': item['timestamp'],
                'asset_id': add_0x_prefix(item['asset-id']),
                'owner': add_0x_prefix(item['owner']),
            })
        return event_list



"""

    Provenance module

"""
import secrets
from typing import Any

PROVENANCE_DEP = 'dep'
PROVENANCE_ACTIVITY_TYPE_PUBLISH = 'publish'
PROVENANCE_ACTIVITY_TYPE_IMPORT = 'import'
PROVENANCE_ACTIVITY_TYPE_INVOKE = 'invoke'

PROVENANCE_AGENT_TYPE_SERVICE_PROVIDER = 'service-provider'
PROVENANCE_ACTIVITY_ID_LENGTH = 32
PROVENANCE_REFERENCE_ID_LENGTH = 32


class Provenance:
    def __init__(self, agent_did: str = None, activity_id: str = None, asset_list: Any = None, inputs_text: str = None):
        """
        :param str agent_did: Agent DID that the invoke was executed on
        :param str activity_id: Activity id or invoke job id that was used to run the invoke, or job to create the asset
        :param list[str] asset_list: List of asset DID's that was used to call the invoke operation
        :param str inputs_text: the JSON string of the input fields

        """

        self._agent_did = agent_did
        self._activity_id = activity_id
        self._asset_list = asset_list
        self._inputs_text = inputs_text
        if self._activity_id is None:
            self._activity_id = self._generate_activity_id()

        self._index_items = {
            'wasAssociatedWith': {
                'prefix': 'assoc_',
                'index': 1,
            },
            'wasGeneratedBy': {
                'prefix': 'gen_',
                'index': 1,
            },
            'wasDerivedFrom': {
                'prefix': 'derived_',
                'index': 1,
            },
        }

    @property
    def create_publish(self) -> Any:
        """

        Return a publish provenance metadata dict. This is called before an asset is published to an agent.

        :param str agent_did: DID of the agent this asset is going to be published too.
        :param str activity_id: Activity id that can be set with this provenance, if not set ,then a random
            value will be set
        """
        return self._generate_publish_import(PROVENANCE_ACTIVITY_TYPE_PUBLISH)

    @property
    def create_import(self) -> Any:
        """

        Return an import provenance records as a dict. This is called before a copied asset is published to an agent.

        :param str agent_did: Optional did of the agent that this asset is being published too.
        :param str activity_id: Activity id that can be set with this provenance, if not set ,then a random
            value will be set
        """
        return self._generate_publish_import(PROVENANCE_ACTIVITY_TYPE_IMPORT)

    @property
    def create_invoke(self) -> Any:

        """

        Create a provenance invoke dict.

        :returns: a dict of the provenance metadata
        """

        entities = self._generate_asset_entity()
        if self._asset_list:
            for asset_did in self._asset_list:
                entities = self._generate_asset_entity(asset_did, entities)

        dependencies = None
        if self._inputs_text:
            dependencies = self._generate_dependencies(self._inputs_text)

        result = {
            'prefix': self._generate_prefix,
            'activity': self._generate_activity(self._activity_id, PROVENANCE_ACTIVITY_TYPE_INVOKE, dependencies),
            'entity': entities,
            'wasGeneratedBy': self._generate_was_generated_by(self._activity_id)
        }
        if self._agent_did:
            result['agent'] = self._generate_agent(self._agent_did, PROVENANCE_AGENT_TYPE_SERVICE_PROVIDER)
            result['wasAssociatedWith'] = self._generate_was_associated_with(self._agent_did, self._activity_id)

        if self._asset_list:
            result['wasDerivedFrom'] = self._generate_was_derived_from(self._asset_list)
        return result

    def _generate_publish_import(self, activity_type: str) -> Any:
        """

        generate a import or publish provenance metadata dict. This is called before an asset is published to an agent.

        :param str agent_did: Optional Agent DID of the agent this asset is going to be published too.
        :param str activity_id: Activity id that can be set with this provenance, if not set ,then a random
            value will be set
        """

        result = {
            'prefix': self._generate_prefix,
            'activity': self._generate_activity(self._activity_id, activity_type),
            'entity': self._generate_asset_entity(),
            'wasGeneratedBy': self._generate_was_generated_by(self._activity_id)
        }

        if self._agent_did:
            result['agent'] = self._generate_agent(self._agent_did, PROVENANCE_AGENT_TYPE_SERVICE_PROVIDER)
            result['wasAssociatedWith'] = self._generate_was_associated_with(self._agent_did, self._activity_id)

        return result

    @property
    def _generate_prefix(self) -> Any:
        return {
            'xsd': 'http://www.w3.org/2001/XMLSchema#',
            'prov': 'http://www.w3.org/ns/prov#',
            PROVENANCE_DEP: 'http://datacraft.sg'
        }

    def _generate_asset_entity(self, asset_did: str = None, entities: Any = None) -> Any:
        if entities is None:
            entities = {}
        if asset_did is None:
            asset_did = f'{PROVENANCE_DEP}:this'
        entities[asset_did] = {
            'prov:type': {
                '$': f'{PROVENANCE_DEP}:asset',
                'type': 'xsd:string'
            }
        }
        return entities

    def _generate_activity(self, activity_id: str, activity_type: str, entries: Any = None) -> Any:
        items = {
            'prov:type': {
                '$': f'{PROVENANCE_DEP}:{activity_type}',
                'type': 'xsd:string'
            }
        }
        if entries:
            for name, value in entries.items():
                items[name] = value

        return {
            f'{activity_id}': items
        }

    def _generate_agent(self, agent_did: str, agent_type: str) -> Any:
        return {
            f'{agent_did}': {
                'prov:type': {
                    '$': f'{PROVENANCE_DEP}:{agent_type}',
                    'type': 'xsd:string'
                }
            }
        }

    def _generate_was_associated_with(self, agent_did: str, activity_id: str) -> Any:
        new_id = self._generate_unique_id('wasAssociatedWith')
        return {
            f'_:{new_id}': {
                'prov:agent': agent_did,
                'prov:activity': activity_id
            }
        }

    def _generate_was_generated_by(self, activity_id: str) -> Any:
        entity_id = f'{PROVENANCE_DEP}:this'
        new_id = self._generate_unique_id('wasGeneratedBy')
        return {
            f'_:{new_id}': {
                'prov:entity': entity_id,
                'prov:activity': activity_id
            }
        }

    def _generate_was_derived_from(self, asset_list: Any) -> Any:
        entities = {}
        for asset_did in asset_list:
            new_id = self._generate_unique_id('wasDerivedFrom')
            entities[f'_:{new_id}'] = {
                'prov:usedEntity': asset_did,
                'prov:generatedEntity': f'{PROVENANCE_DEP}:this'
            }
        return entities

    def _generate_dependencies(self, inputs_text: str) -> Any:
        result = {}
        if inputs_text:
            result[f'{PROVENANCE_DEP}:inputs'] = {
                '$': inputs_text,
                'type': 'xsd:string'
            }
        return result

    def _generate_activity_id(self) -> str:
        return secrets.token_hex(PROVENANCE_ACTIVITY_ID_LENGTH)

    def _generate_unique_id(self, name) -> str:
        item = self._index_items[name]
        new_id = f"{item['prefix']}{item['index']}"
        item['index'] += 1
        return new_id



"""
    Test utils.did

"""
import pytest
import secrets

from starfish.network.did import (
    did_validate,
    did_generate_random,
    did_to_id,
    id_to_did,
    is_did,
    did_parse,
    asset_did_validate,
    decode_to_asset_id,
    is_asset_did
)


def test_did_validate():
    test_did = did_generate_random()
    did_id = did_to_id(test_did)

    did_validate(test_did)
    did_validate(f'{test_did}/test')
    did_validate(f'{test_did}01021A2A/test')

    test_did = (did_id, did_id)
    with pytest.raises(TypeError):
        did_validate(test_did)

    assert(not is_did(test_did))

    test_did = 'test-text'
    with pytest.raises(ValueError, match='must start'):
        did_validate(test_did)
    assert(not is_did(test_did))

    test_did = 'did'
    with pytest.raises(ValueError, match='must start'):
        did_validate(test_did)
    assert(not is_did(test_did))

    test_did = f'did:extra-long:{did_id}'
    with pytest.raises(ValueError, match='"id" must have only'):
        did_validate(test_did)
    assert(not is_did(test_did))

    test_did = f'did:valid:0x01A2B3C'
    with pytest.raises(ValueError, match='path should only have hex characters'):
        did_validate(test_did)
    assert(not is_did(test_did))



def test_did_parse():
    test_did = did_generate_random()
    did_id = did_to_id(test_did)

    # basic parse
    value = did_parse(test_did)

    assert(value['method'] == 'dep')
    assert(f'0x{value["id"]}' == did_id)
    assert(value['id_hex'] == did_id)
    assert(value['fragment'] is None)
    assert(value['path'] is None)


    # test path
    test_path = secrets.token_hex(32)
    value = did_parse(f'{test_did}/{test_path}')

    assert(value['method'] == 'dep')
    assert(f'0x{value["id"]}' == did_id)
    assert(value['id_hex'] == did_id)
    assert(value['fragment'] == '')
    assert(value['path'] == test_path)


    # test fragment
    test_path = secrets.token_hex(32)
    test_fragment = secrets.token_urlsafe(32)
    value = did_parse(f'{test_did}/{test_path}#{test_fragment}')

    assert(value['method'] == 'dep')
    assert(f'0x{value["id"]}' == did_id)
    assert(value['id_hex'] == did_id)
    assert(value['fragment'] == test_fragment)
    assert(value['path'] == test_path)

def test_did_to_id():
    test_did = did_generate_random()
    did_id = did_to_id(test_did)
    value = did_parse(test_did)
    assert(value['id_hex'] == did_id)

def test_id_to_did():
    test_id = secrets.token_hex(32)
    did = id_to_did(test_id)
    value = did_parse(did)
    assert(value['id'] == test_id)

def test_decode_to_asset_id():
    asset_id = secrets.token_hex(32)
    test_did = did_generate_random()

    # decode did/asset_id
    did = f'{test_did}/{asset_id}'
    assert(asset_id == decode_to_asset_id(did))
    assert(is_asset_did(did))
    assert(asset_did_validate(did))

    # decode did/asset_id with leading 0x
    did = f'{test_did}/0x{asset_id}'
    assert(asset_id == decode_to_asset_id(did))
    assert(is_asset_did(did))
    assert(asset_did_validate(did))

    # decode asset_id
    assert(asset_id == decode_to_asset_id(asset_id))

    # decode asset_id with leading 0x
    assert(asset_id == decode_to_asset_id(f'0x{asset_id}'))

    # fail decode with only did
    with pytest.raises(ValueError, match='Unable to get an asset_id'):
        assert(decode_to_asset_id(test_did))

    with pytest.raises(ValueError, match='does not have an asset_id'):
        assert(asset_did_validate(test_did))

    assert(not is_asset_did(test_did))

    # fail decode with invalid asset_id length is too long
    did = f'{test_did}/{asset_id}12A3B4C'
    with pytest.raises(ValueError, match='asset_id is not valid '):
        assert(decode_to_asset_id(did))

    with pytest.raises(ValueError, match='has an invalid asset_id'):
        assert(asset_did_validate(did))

    # passes did valdiaton but fails asset_did validation
    did = f'{test_did}/00112233445566778899AA'
    assert(decode_to_asset_id(did))

    with pytest.raises(ValueError, match='has an invalid asset_id'):
        assert(asset_did_validate(did))

    assert(not is_asset_did(did))




"""

    Memory Asset

"""
import json
import os
from mimetypes import MimeTypes
from typing import (
    Any,
    Generic
)

from starfish.asset.asset_base import AssetBase
from starfish.types import TDataAsset
from starfish.utils.crypto_hash import hash_sha3_256


class DataAsset(AssetBase, Generic[TDataAsset]):
    """

    File asset can be used manage a data asset on the dex network

    :param metadata: Dictionary metadata to provide for the asset.
    :type metadata: None or dict
    :param did: Optional did of the asset if it's registered
    :type did: None or str
    :param str data: Optional data of the asset, this can be str or bytes

    """
    def __init__(self, metadata_text: str, data: Any = None) -> None:

        if data:
            if not (isinstance(data, str) or isinstance(data, bytes)):
                raise ValueError('data can only be str or bytes')
            if not isinstance(data, bytes):
                data = data.encode('utf-8')

        self._data = data
        AssetBase.__init__(self, metadata_text)

    @staticmethod
    def create(name: str, data: Any, metadata: Any = None) -> TDataAsset:
        """

        Create a new DataAsset using string or bytes data.

        :param str name: Name of the asset to create
        :param str, bytes data: Data to assign to the asset
        :param dict metadata: Optional metadata to add to the assets metadata

        :return: a new DataAsset
        :type: :class:`.DataAsset`

        """
        metadata = AssetBase.generateMetadata(name, 'dataset', metadata)

        content_type = 'application/octet-stream'
        if isinstance(data, str):
            content_type = 'text/plain'
        elif isinstance(data, bytes):
            content_type = 'application/octet-stream'
        else:
            raise ValueError('data can only be str or bytes')

        if 'contentType' not in metadata:
            metadata['contentType'] = content_type
        if 'contentHash' not in metadata:
            metadata['contentHash'] = hash_sha3_256(data)

        return DataAsset(json.dumps(metadata), data=data)

    @staticmethod
    def create_from_file(name: str, filename: str, metadata: Any = None, did: str = None, is_read: bool = True) -> TDataAsset:
        """

        Create a new DataAsset using a file or filename.

        :param str name: Name of the asset to create
        :param str filename: If the filename is assigned to a valid file,
            the contents will be saved in the asset
        :param dict metadata: Optional metadata to add to the assets metadata
        :param str did: Option DID to assign to this asset
        :param bool is_read: If True read the file contents in as asset data.

        :return: a new DataAsset
        :type: :class:`.DataAsset`

        """

        metadata = AssetBase.generateMetadata(name, 'dataset', metadata)
        if 'filename' not in metadata:
            metadata['filename'] = os.path.basename(str(filename))
        data = None
        if os.path.exists(filename):
            content_type = 'application/octet-stream'
            mime = MimeTypes()
            mime_type = mime.guess_type(f'file://{filename}')
            if mime_type:
                content_type = mime_type[0]
            if 'contentType' not in metadata:
                metadata['contentType'] = content_type
            if is_read:
                with open(filename, 'rb') as fp:
                    data = fp.read()
                if 'contentLength' not in metadata:
                    metadata['contentLength'] = os.path.getsize(filename)
            if 'contentHash' not in metadata:
                metadata['contentHash'] = hash_sha3_256(data)

        return DataAsset(json.dumps(metadata), data=data)

    def save_to_file(self, filename: str) -> None:
        """
        Saves the data in the data asset to a file.

        :param str filename: Filename to save the data.

        """

        if self._data:
            with open(filename, 'wb') as fp:
                fp.write(self._data)

    @property
    def data(self) -> Any:
        return self._data



"""

    Command Help
"""
from typing import Any

from .command_base import CommandBase


class HelpCommand(CommandBase):

    def __init__(self, sub_parser: Any = None, parent_command: Any = None) -> None:
        self._parent_command = parent_command
        super().__init__('help', sub_parser)

    def create_parser(self, sub_parser: Any) -> Any:

        parser = sub_parser.add_parser(
            self._name,
            description='Get help about this command section',
            help='Get more help'

        )
        return parser

    def execute(self, args: Any, output: Any) -> None:
        self._parent_command.print_help()
